/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */
package org.wso2.registry.app;

import org.apache.abdera.Abdera;
import org.apache.abdera.factory.Factory;
import org.apache.abdera.model.*;
import org.apache.abdera.protocol.Response;
import org.apache.abdera.protocol.client.AbderaClient;
import org.apache.abdera.protocol.client.ClientResponse;
import org.apache.abdera.protocol.client.RequestOptions;
import org.apache.axiom.om.util.Base64;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.registry.Comment;
import org.wso2.registry.*;
import org.wso2.registry.Collection;
import org.wso2.registry.jdbc.handlers.filters.Filter;
import org.wso2.registry.jdbc.handlers.Handler;
import org.wso2.registry.config.RegistryContext;
import org.wso2.registry.exceptions.ResourceNotFoundException;
import org.wso2.registry.i18n.Messages;

import javax.xml.namespace.QName;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.URLConnection;
import java.net.MalformedURLException;
import java.net.URLEncoder;
import java.util.*;

public class RemoteRegistry implements Registry, APPConstants {
    private String baseURI;
    private Log log = LogFactory.getLog(RemoteRegistry.class);
    //This will keep the value of username and the password for authorization
    private String authorizationString = null;
    private String username = null;
    Abdera abdera = new Abdera();

    /**
     * To create a remote registry need to provide a URL of a remote registry and the URL should be
     * something like http://localhost:8080/wso2registry/atom/r1
     *
     * @param registryURL : URL to the registry or to the resource
     */
    public RemoteRegistry(URL registryURL) {
        baseURI = registryURL.toString();
        if (baseURI.endsWith("/")) {
            baseURI = baseURI.substring(0, baseURI.length() - 1);
        }
    }

    /**
     * To create a remote registry to connect to a secure registry or a registry where we need to
     * pass userName and the password to perform any operation. Passing not null value registry will
     * send the username and the password for each request. Need to remember here is that the
     * username and the password send as plain text.
     *
     * @param registryURL : URL to the registry or to the resource
     * @param userName    : user name of the registry , value can be null , if the value is null
     *                    then the user will be the anonymous.
     * @param password    : password , value can be null
     */
    public RemoteRegistry(URL registryURL, String userName, String password) {
        baseURI = registryURL.toString();
        if (baseURI.endsWith("/")) {
            baseURI = baseURI.substring(0, baseURI.length() - 1);
        }
        this.username = userName;
        if (userName != null && password != null) {
            authorizationString = userName + ":" + password;
            authorizationString = "Basic " + Base64.encode(authorizationString.getBytes());
        }
    }

    public RemoteRegistry(String registryURL, String userName, String password)
            throws MalformedURLException {
        this(new URL(registryURL), userName, password);
    }

    public Resource newResource() throws RegistryException {
        ResourceImpl resource = new ResourceImpl();
        resource.setAuthorUserName(username);
        return resource;
    }

    public Collection newCollection() throws RegistryException {
        CollectionImpl collection = new CollectionImpl();
        collection.setAuthorUserName(username);
        return collection;
    }

    public void addHandler(Filter filter, Handler handler) throws RegistryException {
        String msg = "Remote registry does not support adding handlers.";
        log.error(msg);
        throw new RegistryException(msg);
    }

    public Resource get(String path) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse =
                abderaClient.get(baseURI + "/atom" + encodeURL(path), getAuthorization());
        if (clientResponse.getType() == Response.ResponseType.CLIENT_ERROR ||
            clientResponse.getType() == Response.ResponseType.SERVER_ERROR) {
            if (clientResponse.getStatus() == 404)
                throw new ResourceNotFoundException(path);
            throw new RegistryException(clientResponse.getStatusText());
        }
        Element introspection = clientResponse.getDocument().getRoot();
        if (introspection instanceof Feed) {
            // This is a collection
            Feed feed = (Feed)introspection;
            String state = feed.getSimpleExtension(new QName(NAMESPACE, "state"));
            if (state != null && state.equals("Deleted")) {
                throw new ResourceNotFoundException(path);
            }
            return createResourceFromFeed(feed, false);
        }

        Entry entry = (Entry)introspection;
        ResourceImpl resource = createResourceFromEntry(entry);
        resource.setPath(path);
        return resource;
    }

    public Collection get(String path, int start, int pageSize) throws RegistryException {
        // todo - implement me
        return null;
    }

    /**
     * This method will generate a resource object representing the Feed object and the logic will
     * be simply the reverse of the create feed of the Atom registry
     *
     * @param feed    : Feed object which represent a resource object
     * @param isQuery : To tell whether we are going to execute a query of trying to just get a
     *                resource
     * @return : Created resource
     * @throws org.wso2.registry.RegistryException
     *          : If user is unable to open the URL connection
     */
    private ResourceImpl createResourceFromFeed(Feed feed, boolean isQuery)
            throws RegistryException {
        CollectionImpl resource = new CollectionImpl();
        Properties properties = feed.getExtension(PropertyExtensionFactory.PROPERTIES);
        RemoteRegistry.createPropertiesFromExtensionElement(properties, resource);
        if (feed.getAuthor() != null) {
            resource.setAuthorUserName(feed.getAuthor().getName());
        }
        resource.setLastModified(feed.getUpdated());
        String createdDate = feed.getSimpleExtension(new QName(NAMESPACE, "createdTime"));
        if (createdDate != null) {
            resource.setCreatedTime(new Date(Long.parseLong(createdDate)));
        }

        String lastUpdatedUser = feed.getSimpleExtension(APPConstants.QNAME_LAST_UPDATER);
        if (lastUpdatedUser != null) {
            resource.setLastUpdaterUserName(lastUpdatedUser);
        }
        String mediaType = feed.getSimpleExtension(new QName(NAMESPACE, "mediaType"));
        if (mediaType != null) {
            resource.setMediaType(mediaType);
        }
        final Link pathLink = feed.getLink("path");
        String path = (pathLink != null) ? pathLink.getHref().toString() : feed.getTitle();
        resource.setPath(path);
        resource.setDescription(feed.getSubtitle());
        String state = feed.getSimpleExtension(new QName(NAMESPACE, "state"));
        if (state != null && "Deleted".equals(state)) {
            resource.setState(RegistryConstants.DELETED_STATE);
        }

        String isComments = feed.getSimpleExtension(COMMENTS_QNAME);
        if (isComments != null) {
            resource.setContent(getCommentsFromFeed(feed));
        } else {
            List entries = feed.getEntries();
            if (entries != null) {
                String[] childNodes = new String[entries.size()];
                for (int i = 0; i < entries.size(); i++) {
                    Entry entry = (Entry)entries.get(i);
                    if (isQuery) {
                        childNodes[i] = entry.getLink("queryResult").getHref().toString();
                    } else {
                        childNodes[i] = entry.getLink("path").getHref().toString();
                    }

                }
                resource.setContent(childNodes);
            }
        }

        return resource;
    }

    private ResourceImpl createResourceFromEntry(Entry entry) throws RegistryException {
        ResourceImpl resource = new ResourceImpl();

        // TODO -
        resource.setPath(entry.getId().toString());
        String mediaType = entry.getSimpleExtension(new QName(NAMESPACE, "mediaType"));
        if (mediaType != null) {
            resource.setMediaType(mediaType);
        }

        Properties properties = entry.getExtension(PropertyExtensionFactory.PROPERTIES);
        RemoteRegistry.createPropertiesFromExtensionElement(properties, resource);
        if (entry.getAuthor() != null) {
            resource.setAuthorUserName(entry.getAuthor().getName());
        }
        resource.setLastModified(entry.getUpdated());
        String createdDate = entry.getSimpleExtension(new QName(NAMESPACE, "createdTime"));
        if (createdDate != null) {
            resource.setCreatedTime(new Date(Long.parseLong(createdDate)));
        }

        String lastUpdatedUser = entry.getSimpleExtension(APPConstants.QNAME_LAST_UPDATER);
        if (lastUpdatedUser != null) {
            resource.setLastUpdaterUserName(lastUpdatedUser);
        }
        
        // TODO - get content-type in an Atom-appropriate way
//        String mediaType = entry.getSimpleExtension(new QName(NAMESPACE, "mediaType"));

        resource.setDescription(entry.getSummary());

        Link link = entry.getLink("edit-media");
        if (link != null) {
            try {
                URL url = new URL(new URL(baseURI), link.getHref().toString());
                URLConnection uc = url.openConnection();
                if (authorizationString != null) {
                    uc.setRequestProperty("Authorization", authorizationString);
                }
                InputStream in = url.openStream();
                resource.setContentStream(in);
                return resource;
            } catch (IOException e) {
                throw new RegistryException(Messages.getMessage("unable.to.receive",
                                                                link.getHref().toString()));
            }
        }
        String content1 = entry.getContent();
        resource.setContent(content1);
        return resource;
    }

    public boolean resourceExists(String path) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse response = abderaClient.head(baseURI + APPConstants.ATOM + encodeURL(path),
                                                    getAuthorization());
        return (response.getType() == Response.ResponseType.SUCCESS);
    }

    public String put(String suggestedPath, Resource resource) throws RegistryException {
        // Does the resource already exist?  If so this is an update (PUT) not a create (POST)
        boolean alreadyExists = resourceExists(suggestedPath);

        int idx = suggestedPath.lastIndexOf("/");
        String relativePath = suggestedPath.substring(idx + 1);
        String parentPath = idx > 1 ? suggestedPath.substring(0, idx) : "/";

        AbderaClient abderaClient = new AbderaClient(abdera);
        final Factory factory = abdera.getFactory();
        boolean isCollection = resource instanceof Collection;

        ExtensibleElement element;
        if (isCollection) {
            Feed feed = factory.newFeed();
            feed.setId(baseURI + APPConstants.ATOM + encodeURL(suggestedPath));
//            feed.setId(encodeURL(suggestedPath));
            feed.setTitle(suggestedPath);
            feed.setSubtitle(resource.getDescription());
            feed.addAuthor(username);
            feed.setUpdated(new Date());
            element = feed;
        } else {
            Entry entry = factory.newEntry();
            entry.setId(baseURI + APPConstants.ATOM + encodeURL(suggestedPath));
//            entry.setId(encodeURL(suggestedPath));
            entry.setTitle(suggestedPath);
            entry.setSummary(resource.getDescription());
            entry.addAuthor(username);
            entry.setUpdated(new Date());
            Object content = resource.getContent();
            if (content != null && content instanceof byte[]) {
                ByteArrayInputStream in = new ByteArrayInputStream((byte[])content);
                entry.setContent(in);
            } else if (content instanceof InputStream) {
                entry.setContent((InputStream)content);
            } else {
                entry.setContent((String)content);
            }
            element = entry;
        }
        java.util.Properties properties = resource.getProperties();
        addPropertyExtensionElement(properties, factory, element,
                                    PropertyExtensionFactory.PROPERTIES,
                                    PropertyExtensionFactory.PROPERTY);
        final String mediaType = resource.getMediaType();
        if (mediaType != null && mediaType.length() > 0) {
            element.addSimpleExtension(new QName(NAMESPACE, "mediaType"), mediaType);
        }
        element.addSimpleExtension(new QName(NAMESPACE, "parentPath"), resource.getParentPath());
        if (((ResourceImpl)resource).isContentModified()) {
            element.addSimpleExtension(new QName(NAMESPACE, "contentModified"), "true");
        }

        RequestOptions requestOptions = getAuthorization();
        requestOptions.setSlug(relativePath);

        ClientResponse resp;
        if (!alreadyExists) {
            resp = abderaClient.post(baseURI + APPConstants.ATOM + encodeURL(parentPath),
                                     element, requestOptions);
        } else {
            resp = abderaClient.put(baseURI + APPConstants.ATOM + encodeURL(suggestedPath),
                                    element, requestOptions);
        }
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.add", suggestedPath));
        } else {
            log.error(Messages.getMessage("add.resource.fail", suggestedPath));
            throw new RegistryException(Messages.getMessage("add.resource.fail", suggestedPath));
        }
        String resultPath = suggestedPath;
        if (!alreadyExists) {
            resultPath = resp.getLocation().getASCIIPath().substring("/wso2registry/atom".length());
        }
//        ResourceImpl impl = (ResourceImpl)resource;
//        impl.setPath(resultPath);
//        // todo - fix this to use util routine?
//        int i = resultPath.lastIndexOf('/');
//        impl.setParentPath(i == 0 ? "/" : resultPath.substring(0, i));
        return resultPath;
    }

    public static void createPropertiesFromExtensionElement(Properties properties,
                                                            Resource resource) {
        if (properties != null) {
            List propertyList = properties.getExtensions(PropertyExtensionFactory.PROPERTY);
            for (Object aPropertyList : propertyList) {
                Property property = (Property)aPropertyList;
                PropertyName pn = property.getExtension(PropertyExtensionFactory.PROPERTY_NAME);
                PropertyValue pv = property.getExtension(PropertyExtensionFactory.PROPERTY_VALUE);
                String value = pv.getText();
                List list =  decodeMultivaluedProperty(value);
                for (Object aList : list) {
                    String stringValue = (String)aList;
                    resource.addProperty(pn.getText(), stringValue);
                }
            }
        }
    }

    /**
     * This will generate extension element and that will add to the entry , the created element
     * will be something like <node1name> <node2Name> + <name></name> <value></value> </node2Name>
     * </node1name>
     *
     * @param properties : List of Name value pairs
     * @param factory    : Abdera Factory
     * @param entry      : Instanc of entery where extension element need to add
     * @param node1name  : Name of node 1
     * @param node2Name  : Name of node 2
     */
    public static void addPropertyExtensionElement(java.util.Properties properties,
                                                   Factory factory,
                                                   ExtensibleElement entry,
                                                   QName node1name,
                                                   QName node2Name) {
        if (properties != null) {
            Properties propertyElement = factory.newExtensionElement(node1name);
            Iterator keys = properties.keySet().iterator();
            while (keys.hasNext()) {
                String key = (String)keys.next();
                Property property = factory.newExtensionElement(node2Name);
                PropertyName pn = factory.newExtensionElement(
                        PropertyExtensionFactory.PROPERTY_NAME);
                pn.setPropertyName(key);
                PropertyValue pv = factory.newExtensionElement(
                        PropertyExtensionFactory.PROPERTY_VALUE);
                Object valueList = properties.get(key);
                String propertyValue ="" ;
                if (valueList instanceof List) {
                    propertyValue = encodeMultivaluedProperty((List)valueList, propertyValue);
                    pv.setPropertyValue(propertyValue);
                    property.setProperty(pn, pv);
                    propertyElement.setProperty(property);
                }
            }
            entry.addExtension(propertyElement);
        }
    }

    private static String encodeMultivaluedProperty(List valueList, String propertyValue) {
        for (Object aValueList : valueList) {
            String value = (String)aValueList;
            if ("".equals(propertyValue)) {
                propertyValue = value;
            } else {
                propertyValue = propertyValue + "|" + value;
            }
        }
        return propertyValue;
    }

    public static List decodeMultivaluedProperty(String propertyValue) {
        String values [] = propertyValue.split("\\|");
        List list = new ArrayList();
        list.addAll(Arrays.asList(values));
        return list;
    }


    public String importResource(String suggestedPath, String sourceURL, Resource metadata)
            throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);

        String postData = "?importURL=" + sourceURL;
        ByteArrayInputStream bis = new ByteArrayInputStream(postData.getBytes());
        RequestOptions opts = getAuthorization();
        opts.setSlug(suggestedPath);
        opts.setContentType(IMPORT_MEDIATYPE);
        ClientResponse response =
                abderaClient.post(baseURI + APPConstants.ATOM + encodeURL(postData),
                                  bis,
                                  opts);
        return response.getLocation().toString();
    }

    public void delete(String path) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse resp = abderaClient.delete(baseURI + APPConstants.ATOM + encodeURL(path),
                                                  getAuthorization());
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.deleted", path));
        } else {
            log.error(Messages.getMessage("resource.delete.fail", path));
            throw new RegistryException(Messages.getMessage("resource.delete.fail", path));
        }
    }


    public String rename(String currentPath, String newPath) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        ByteArrayInputStream is = new ByteArrayInputStream(newPath.getBytes());
        ClientResponse resp =
                abderaClient.post(baseURI + APPConstants.ATOM +
                                  encodeURL(currentPath +
                                            RegistryConstants.URL_SEPARATOR +
                                            PARAMETER_RENAME),
                                  is,
                                  getAuthorization());
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.rename", currentPath));
        } else {
            log.error(Messages.getMessage("rename.resource.fail", currentPath));
            throw new RegistryException(Messages.getMessage("rename.resource.fail", currentPath));
        }
        return newPath;
    }

    public String move(String currentPath, String newPath) throws RegistryException {
        // todo: provide APP support for move
        return null;
    }

    public String copy(String sourcePath, String targetPath) throws RegistryException {
        // todo: provide APP support for copy
        return null;
    }

    public void createVersion(String path) throws RegistryException {
    }

    public String[] getVersions(String path) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse =
                abderaClient.get(baseURI + APPConstants.ATOM +
                                 encodeURL(path +
                                           RegistryConstants.URL_SEPARATOR +
                                           PARAMETER_VERSION),
                                 getAuthorization());
        Document introspection = clientResponse.getDocument();
        Feed feed = (Feed)introspection.getRoot();
        List entries = feed.getEntries();
        if (entries != null) {
            String[] versions = new String[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entry = (Entry)entries.get(i);
                versions[i] = entry.getLink("versionLink").getHref().toString();
            }
            return versions;
        }
        return new String[0];
    }

    public void restoreVersion(String versionPath) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        ClientResponse resp = abderaClient.post(baseURI + APPConstants.ATOM +
                                                encodeURL(versionPath +
                                                          RegistryConstants.URL_SEPARATOR +
                                                          PARAMETER_RESTORE),
                                                entry,
                                                getAuthorization());
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.restoreded", versionPath));
        } else {
            log.error(Messages.getMessage("resource.restore.fail", versionPath));
            throw new RegistryException(Messages.getMessage("resource.restore.fail", versionPath));
        }
    }

    public void addAssociation(String associationType, String sourcePath, String associationPaths)
            throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        final Factory factory = abdera.getFactory();
        Element el = factory.newElement(QNAME_ASSOC);
        el.setAttributeValue(ASSOC_TYPE, associationType);
        el.setText(associationPaths);
        ClientResponse resp = abderaClient.post(baseURI + APPConstants.ATOM +
                                                encodeURL(sourcePath +
                                                          RegistryConstants.URL_SEPARATOR +
                                                          APPConstants.ASSOCIATIONS),
                                                el,
                                                getAuthorization());
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.associated", sourcePath));
        } else {
            log.error(Messages.getMessage("resource.associated.fails", sourcePath));
            throw new RegistryException(Messages.getMessage("resource.tag.fails", sourcePath));
        }
    }


    public void removeAssociation(String associationType, String sourcePath, String associationPaths) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        final Factory factory = abdera.getFactory();
        Element el = factory.newElement(QNAME_ASSOC);
        el.setAttributeValue(ASSOC_TYPE, associationType);
        el.setText(associationPaths);
        ClientResponse resp = abderaClient.put(baseURI + APPConstants.ATOM +
                                                encodeURL(sourcePath +
                                                          RegistryConstants.URL_SEPARATOR +
                                                          PARAMETER_REMOVE_ASSOCIATION),
                                                el,
                                                getAuthorization());
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.associated", sourcePath));
        } else {
            log.error(Messages.getMessage("resource.associated.fails", sourcePath));
            throw new RegistryException(Messages.getMessage("resource.tag.fails", sourcePath));
        }
    }

    public Association[] getAllAssociations(String resourcePath) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(baseURI + APPConstants.ATOM +
                                                         encodeURL(resourcePath +
                                                                   RegistryConstants.URL_SEPARATOR +
                                                                   ASSOCIATIONS),
                                                         getAuthorization());
        Document introspection = clientResponse.getDocument();
        Feed feed = (Feed)introspection.getRoot();
        return getAssociationsFromFeed(feed, resourcePath);
    }

    public static Association [] getAssociationsFromFeed(Feed feed, String resourcePath) {
        List entries = feed.getEntries();
        Association associations[] = null;
        if (entries != null) {
            associations = new Association[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entry = (Entry)entries.get(i);
                Association association = new Association();
                association.setSourcePath(resourcePath);
                association.setDestinationPath(entry.getContent());
                association.setAssociationType(entry.getSummary());
                associations[i] = association;
            }
        }
        return associations;
    }

    public Association[] getAssociations(String resourcePath, String associationType) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse =
                abderaClient.get(baseURI + APPConstants.ATOM +
                                 encodeURL(resourcePath + RegistryConstants.URL_SEPARATOR +
                                           ASSOCIATIONS + ":" +
                                           associationType),
                                 getAuthorization());
        Document introspection = clientResponse.getDocument();
        Feed feed = (Feed)introspection.getRoot();
        List entries = feed.getEntries();
        Association associations[] = null;
        if (entries != null) {
            associations = new Association[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entry = (Entry)entries.get(i);
                Association association = new Association();
                association.setSourcePath(resourcePath);
                association.setDestinationPath(entry.getContent());
                association.setAssociationType(associationType);
                associations[i] = association;
            }
        }
        return associations;
    }

    public void applyTag(String resourcePath, String tag) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.setContent(tag);
        ClientResponse resp = abderaClient.post(baseURI + APPConstants.ATOM +
                                                encodeURL(resourcePath +
                                                          RegistryConstants.URL_SEPARATOR +
                                                          PARAMETER_TAGS),
                                                entry,
                                                getAuthorization());
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.debug(Messages.getMessage("resource.tagged", resourcePath));
        } else {
            log.error(Messages.getMessage("resource.tag.fails", resourcePath));
            throw new RegistryException(Messages.getMessage("resource.tag.fails", resourcePath));
        }
    }

    public TaggedResourcePath[] getResourcePathsWithTag(String tag) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(baseURI + "/tags/" + tag,
                                                         getAuthorization());
        Document introspection =
                clientResponse.getDocument();
        Feed feed = (Feed)introspection.getRoot();
        List entries = feed.getEntries();
        TaggedResourcePath taggedResourcePaths[] = null;
        if (entries != null) {
            taggedResourcePaths = new TaggedResourcePath[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entry = (Entry)entries.get(i);
                Properties properties = entry.getExtension(PropertyExtensionFactory.TAGS);
                List propertyList = properties.getExtensions(PropertyExtensionFactory.TAG);
                Map map = new HashMap();
                for (Object aPropertyList : propertyList) {
                    Property property = (Property)aPropertyList;
                    PropertyName pn = property.getExtension(PropertyExtensionFactory.PROPERTY_NAME);
                    PropertyValue pv =
                            property.getExtension(PropertyExtensionFactory.PROPERTY_VALUE);
                    map.put(pn.getText(), pv.getText());
                }
                TaggedResourcePath tagpath = new TaggedResourcePath();
                tagpath.setResourcePath(entry.getTitle());
                tagpath.setTagCount(
                        Long.parseLong(entry.getSimpleExtension(new QName(NAMESPACE,
                                                                          "taggings"))));
                taggedResourcePaths[i] = tagpath;
                tagpath.setTagCounts(map);
            }
        }
        return taggedResourcePaths;
    }

    public Tag[] getTags(String resourcePath) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(baseURI + APPConstants.ATOM +
                                                         encodeURL(resourcePath +
                                                                   RegistryConstants.URL_SEPARATOR +
                                                                   PARAMETER_TAGS),
                                                         getAuthorization());
        Document introspection =
                clientResponse.getDocument();
        Feed feed = (Feed)introspection.getRoot();
        List entries = feed.getEntries();
        Tag tags[] = null;
        if (entries != null) {
            tags = new Tag[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entry = (Entry)entries.get(i);
                Tag tag = new Tag();
                tag.setTagCount(Long.parseLong(entry.getSimpleExtension(new QName(NAMESPACE,
                                                                                  "taggings"))));
                tag.setTagName(entry.getTitle());
                tags[i] = tag;
            }
        }
        return tags;
    }

    public void removeTag(String path, String tag) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse resp = abderaClient.delete(baseURI + APPConstants.ATOM +
                                                  encodeURL(path +
                                                            RegistryConstants.URL_SEPARATOR +
                                                            "tag:" + tag),
                                                  getAuthorization());
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("tag.removed", path));
        } else {
            log.error(Messages.getMessage("tag.removed.fails", path));
            throw new RegistryException(Messages.getMessage("tag.removed.fails", path));
        }
    }

    public String addComment(String resourcePath, Comment comment) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.setId("tag:commentID"); // TODO - generate real ID
        entry.setTitle("Comment");
        entry.setUpdated(comment.getTime());
        entry.addAuthor(comment.getUser());
        entry.setContent(comment.getText());
        ClientResponse resp =
                abderaClient.post(baseURI + APPConstants.ATOM +
                                  encodeURL(resourcePath +
                                            RegistryConstants.URL_SEPARATOR +
                                            PARAMETER_COMMENTS),
                                  entry,
                                  getAuthorization());
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.commented", resourcePath));
            return resp.getLocation().toString().substring(baseURI.length());
        } else {
            log.error(Messages.getMessage("resouce.comment.fails", resourcePath));
            throw new RegistryException(Messages.getMessage("resouce.comment.fails", resourcePath));
        }
    }


    public void editComment(String commentPath, String text) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.setContent(text);
        ClientResponse resp = abderaClient.put(baseURI + APPConstants.ATOM +
                                               encodeURL(commentPath),
                                               entry,
                                               getAuthorization());
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.commented", commentPath));
        } else {
            log.error(Messages.getMessage("resource.comment.fails", commentPath));
            throw new RegistryException(Messages.getMessage("resource.comment.fails", commentPath));
        }
    }

    public Comment[] getComments(String resourcePath) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse =
                abderaClient.get(baseURI + APPConstants.ATOM +
                                 encodeURL(resourcePath +
                                           RegistryConstants.URL_SEPARATOR +
                                           PARAMETER_COMMENTS),
                                 getAuthorization());
        Document introspection = clientResponse.getDocument();
        Element element = introspection.getRoot();
        Feed feed = (Feed)element;
        return getCommentsFromFeed(feed);
    }

    private Comment[] getCommentsFromFeed(Feed feed) {
        List entries = feed.getEntries();
        Comment comments[] = null;
        if (entries != null) {
            comments = new Comment[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entry = (Entry)entries.get(i);
                Comment comment = new Comment();
                comment.setTime(entry.getUpdated());
                final Link resourceLink = entry.getLink("resourcePath");
                if (resourceLink != null) {
                    comment.setResourcePath(resourceLink.getHref().toString());
                }
                if (entry.getLink("direct") != null) {
                    comment.setPath(entry.getLink("direct").getHref().toString());
                }
                comment.setPath(entry.getId().toString());
                comment.setText(entry.getContent());
                comment.setUser(entry.getAuthor().getName());
                comments[i] = comment;
            }
        }
        return comments;
    }

    public void rateResource(String resourcePath, int rating) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        ByteArrayInputStream is = new ByteArrayInputStream(Integer.toString(rating).getBytes());
        ClientResponse resp = abderaClient.post(baseURI + APPConstants.ATOM +
                                                encodeURL(resourcePath +
                                                          RegistryConstants.URL_SEPARATOR +
                                                          PARAMETER_RATINGS),
                                                is,
                                                getAuthorization());
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.rated", resourcePath));
        } else {
            log.error(Messages.getMessage("resource.rate.fail", resourcePath));
            throw new RegistryException(Messages.getMessage("resource.rate.fail", resourcePath));
        }
    }

    public float getAverageRating(String resourcePath) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse =
                abderaClient.get(baseURI + APPConstants.ATOM +
                                 encodeURL(resourcePath +
                                           RegistryConstants.URL_SEPARATOR +
                                           PARAMETER_RATINGS),
                                 getAuthorization());

// TODO!!!  Otherwise we get a ClassCastException below...
//        if (clientResponse.getStatus() != 200) {
//            // throw RegistryException
//        }

        Document introspection = clientResponse.getDocument();
        Feed feed = (Feed)introspection.getRoot();
        String floatvalue = feed.getSimpleExtension(QNAME_AVGRATING);
        return Float.parseFloat(floatvalue);
    }

    public int getRating(String path, String userName) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse =
                abderaClient.get(baseURI + APPConstants.ATOM +
                                 encodeURL(path + RegistryConstants.URL_SEPARATOR +
                                           PARAMETER_RATINGS + ":" + userName),
                                 getAuthorization());
        Document introspection =
                clientResponse.getDocument();
        Feed feed = (Feed)introspection.getRoot();
        String intvalue = feed.getSubtitle();
        return Integer.parseInt(intvalue);
    }

    public Collection executeQuery(String path, Map parameters) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        RequestOptions requestOptions = getAuthorization();
        ClientResponse resp = abderaClient.get(baseURI + APPConstants.ATOM +
                                               encodeURL(path +
                                                         "?" +
                                                         buildQueryString(parameters)),
                                               requestOptions);
        Document introspection = resp.getDocument();
        Feed feed = (Feed)introspection.getRoot();
        return (Collection)createResourceFromFeed(feed, true);
    }

    public static String buildQueryString(Map parameters) {
        StringBuffer buffer = new StringBuffer();
        boolean first = true;
        Iterator iParams = parameters.keySet().iterator();
        while (iParams.hasNext()) {
            String name = (String)iParams.next();
            String value = (String)parameters.get(name);
            if (!first) {
                buffer.append("&");
            } else {
                first = false;
            }
            try {
                buffer.append(URLEncoder.encode(name, "utf-8"));
                buffer.append("=");
                buffer.append(URLEncoder.encode(value, "utf-8"));
            } catch (UnsupportedEncodingException e) {
                // What the heck?? :)
                return "";
            }
        }
        return buffer.toString();
    }

    public static Map decodeQueryString(String value) {

        Map paramMap = new HashMap();

        String[] params = value.trim().split("&");
        for (String param : params) {
            if (param.trim().length() == 0) {
                continue;
            }

            String[] paramParts = param.trim().split("=");
            paramMap.put(paramParts[0], paramParts[1]);
        }

        return paramMap;
    }

    public LogEntry[] getLogs(String resourcePath,
                              int action,
                              String userName,
                              Date from,
                              Date to,
                              boolean recentFirst) throws RegistryException {

        if (resourcePath == null || "".equals(resourcePath)) {
            resourcePath = "/";
        }
        AbderaClient abderaClient = new AbderaClient(abdera);
        RequestOptions requestOptions = getAuthorization();
        requestOptions.addDateHeader("ToDate", to);
        requestOptions.addDateHeader("FromDate", from);
        requestOptions.addHeader("Action", "" + action);
        requestOptions.addHeader("Author", userName);
        ClientResponse resp = abderaClient.get(baseURI + APPConstants.ATOM +
                                               encodeURL(resourcePath +
                                                         RegistryConstants.URL_SEPARATOR +
                                                         PARAMETER_LOGS),
                                               requestOptions);
        Document introspection =
                resp.getDocument();
        Feed feed = (Feed)introspection.getRoot();
        List entries = feed.getEntries();
        LogEntry logs[] = null;
        if (entries != null) {
            logs = new LogEntry[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entry = (Entry)entries.get(i);
                LogEntry logEntry = new LogEntry();
                logEntry.setDate(entry.getEdited());
                logEntry.setActionData(entry.getContent());
                logEntry.setUserName(entry.getAuthor().getName());
                logEntry.setAction(Integer.parseInt(
                        entry.getSimpleExtension(new QName(NAMESPACE, "action"))));
                logEntry.setResourcePath(entry.getSimpleExtension(new QName(NAMESPACE, "path")));
                logs[i] = logEntry;
            }
        }
        return logs;
    }

    /**
     * This method will create a RequestOptions object adding  Authorization headers.
     *
     * @return RequestOptions : Created RequestOptions object
     */
    private RequestOptions getAuthorization() {
        RequestOptions requestOptions = new RequestOptions();
        requestOptions.setAuthorization(authorizationString);
        return requestOptions;
    }

    public static String encodeURL(String path) {
        //I know I should use URLEncoder but when I used that it tries
        //  to encode "/" as well which is not need here
        return path.replaceAll(" ", "+");
    }

    public void addUser(String userName ,
                        String password ,
                        String confirmpassword ,
                        String roleName ,
                        String friendlyName) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.addSimpleExtension(new QName(NAMESPACE, "username"), userName);
        entry.addSimpleExtension(new QName(NAMESPACE, "password"), password);
        entry.addSimpleExtension(new QName(NAMESPACE, "confirmpassword"), confirmpassword);
        entry.addSimpleExtension(new QName(NAMESPACE, "rolename"), roleName);
        entry.addSimpleExtension(new QName(NAMESPACE, "friendlyName"), friendlyName);

        abderaClient.post(baseURI + APPConstants.ATOM + encodeURL(RegistryConstants.URL_SEPARATOR +
                                                                  PARAMETER_ADD_USER),
                          entry,
                          getAuthorization());
    }

    public void addAspect(String name, Aspect aspect) throws RegistryException {
        throw new UnsupportedOperationException();
    }

    public void associateAspect(String resourcePath, String aspect) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        // POST as text to "<resource>;aspects"
        ByteArrayInputStream is = new ByteArrayInputStream(aspect.getBytes());
        ClientResponse resp = abderaClient.post(baseURI + APPConstants.ATOM +
                                                encodeURL(resourcePath +
                                                          RegistryConstants.URL_SEPARATOR +
                                                          ASPECTS),
                                                is,
                                                getAuthorization());
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.associated", resourcePath , aspect));
        } else {
            log.error(Messages.getMessage("resource.associated.fail", resourcePath , aspect));
            throw new RegistryException(Messages.getMessage("resource.associated", resourcePath , aspect));
        }
    }

    public void invokeAspect(String resourcePath, String aspectName, String action) throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);

        // The content doesn't really matter here, so this is a placeholder for now.
        // Later on we'll likely want to support parameterized invocations, so we'll likely
        // enable posting form-encoded data.
        ByteArrayInputStream is = new ByteArrayInputStream("invoke".getBytes());
        ClientResponse resp = abderaClient.post(baseURI + APPConstants.ATOM +
                                                encodeURL(resourcePath +
                                                          RegistryConstants.URL_SEPARATOR +
                                                          ASPECT + "(" + aspectName + ")" +
                                                          action),
                                                is,
                                                getAuthorization());
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("invokeAspect", resourcePath , aspectName , action));
        } else {
            log.error(Messages.getMessage("resource.associated.fail", resourcePath , aspectName , action));
            throw new RegistryException(Messages.getMessage("resource.associated", resourcePath , aspectName , action));
        }
    }

    public void beginTransaction() throws RegistryException {
        //TODO: Method implementation

    }

    public void commitTransaction() throws RegistryException {
        //TODO: Method implementation

    }

    public void rollbackTransaction() throws RegistryException {
        //TODO: Method implementation

    }

    public String[] getAspectActions(String resourcePath, String aspectName)
            throws RegistryException {
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse =
                abderaClient.get(baseURI + APPConstants.ATOM +
                                 encodeURL(resourcePath + RegistryConstants.URL_SEPARATOR +
                                           ASPECTS + ":" + aspectName),
                                 getAuthorization());
        Document introspection = clientResponse.getDocument();
        Feed feed = (Feed)introspection.getRoot();
        List entries = feed.getEntries();
        if (entries != null) {
            String[] aspectActions = new String[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entry = (Entry)entries.get(i);
                aspectActions[i] = entry.getLink("aspectActionsLink").getHref().toString();
            }
            return aspectActions;
        }
        return new String[0];
    }

    public RegistryContext getRegistryContext() {
        return null;
    }

    public void setRegistryContext(RegistryContext registryContext) {
    }
}
