/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc;

import org.wso2.registry.*;
import org.wso2.registry.Collection;
import org.wso2.registry.users.UserRealm;
import org.wso2.registry.users.UserStoreException;
import org.wso2.registry.config.RegistryContext;
import org.wso2.registry.i18n.Messages;
import org.wso2.registry.utils.RegistryUtils;
import org.wso2.registry.utils.VersionedPath;
import org.wso2.registry.session.CurrentSession;
import org.wso2.registry.exceptions.ResourceNotFoundException;
import org.wso2.registry.jdbc.handlers.HandlerManager;
import org.wso2.registry.jdbc.handlers.RequestContext;
import org.wso2.registry.jdbc.handlers.Handler;
import org.wso2.registry.jdbc.handlers.filters.Filter;
import org.wso2.registry.jdbc.queries.QueryProcessorManager;
import org.wso2.registry.jdbc.dao.*;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.sql.DataSource;
import java.util.*;

public class TransactionalJDBCRegistry implements Registry {

    public static final Log log = LogFactory.getLog(TransactionalJDBCRegistry.class);

    private HandlerManager handlerManager;
    private Repository repository;
    private QueryProcessorManager queryProcessorManager;
    private Map<String, Aspect> aspects;
    private VersionRepository versionRepository;

    private ResourceDAO resourceDAO = new ResourceDAO();
    private AssociationDAO associationDAO = new AssociationDAO();
    private TagsDAO tagsDAO = new TagsDAO();
    private CommentsDAO commentsDAO = new CommentsDAO();
    private RatingsDAO ratingsDAO = new RatingsDAO();
    private LogsDAO logsDAO = new LogsDAO();

    public TransactionalJDBCRegistry(
            HandlerManager handlerManager,
            Repository repository,
            VersionRepository versionRepository,
            QueryProcessorManager queryProcessorManager,
            Map<String, Aspect> aspects) {

        this.handlerManager = handlerManager;
        this.repository = repository;
        this.versionRepository = versionRepository;
        this.queryProcessorManager = queryProcessorManager;
        this.aspects = aspects;
    }

    public void beginTransaction() throws RegistryException {
    }

    public void commitTransaction() throws RegistryException {
    }

    public void rollbackTransaction() throws RegistryException {
    }

    public Resource newResource() throws RegistryException {
        String msg = "Transactional registry cannot create resource instances. Please use " +
                "JDBCRegistry for instantiating new resource and collection instances.";
        log.error(msg);
        throw new RegistryException(msg);
    }

    public Collection newCollection() throws RegistryException {
        String msg = "Transactional registry cannot create collection instances. Please use " +
                "JDBCRegistry for instantiating new resource and collection instances.";
        log.error(msg);
        throw new RegistryException(msg);
    }

    public RegistryContext getRegistryContext() {
        return null;
    }

    public void setRegistryContext(RegistryContext registryContext) {
    }

    public void addHandler(Filter filter, Handler handler) throws RegistryException {
        String msg = "Transactional registry does not support adding handlers.";
        log.error(msg);
        throw new RegistryException(msg);
    }

    /**
     * Returns the resource in the given path.
     *
     * @param path Path of the resource. e.g. /project1/server/deployment.xml
     * @return Resource instance
     * @throws org.wso2.registry.RegistryException ResourceNotFound exception is thrown if the resource is not in the
     *                           registry
     */
    public Resource get(String path) throws RegistryException {

        path = RegistryUtils.prepareGeneralPath(path);

        // check if this path refers to a resource referred by a URL query (e.g. comment)

        RequestContext requestContext = new RequestContext(this);
        requestContext.setResourcePath(path);

        Resource resource = handlerManager.get(requestContext);

        // resource may have been fetched from the repository, to be used by handlers. if it is
        // done, it has to be stored in the request context. we can just use that resource
        // without fetching it again from the repository.
        if (resource == null) {
            resource = requestContext.getResource();
        }

        if (resource == null) {
            VersionedPath versionedPath = RegistryUtils.getVersionedPath(path);
            if (versionedPath.getVersion() == -1) {
                resource = repository.get(path);
            } else {
                resource = versionRepository.get(versionedPath);
            }
        }

        if (resource == null) {
            throw new ResourceNotFoundException(path);
        }

        // if the request is for the current version of the resource, it should not be in the
        // deleted state.
        if (path.indexOf("?v=") == -1 && resource.getState() == RegistryConstants.DELETED_STATE) {
            throw new ResourceNotFoundException(path);
        }

        return resource;
    }

    public Collection get(String path, int start, int pageSize) throws RegistryException {
        return repository.get(path, start, pageSize);
    }

    public boolean resourceExists(String path) throws RegistryException {

        VersionedPath versionedPath = RegistryUtils.getVersionedPath(path);

        if (versionedPath.getVersion() == -1) {
            return repository.resourceExists(path);
        } else {
            // todo: delegate to version repository
            return false;
        }
    }

    /**
     * Adds or updates resources in the registry. If there is no resource in the given path,
     * resource is added. If a resource already exist in the given path, it will be replaced with
     * the new resource. Implementations may use a concurrency control method to prevent concurrency
     * issues.
     *
     * @param suggestedPath the path which we'd like to use for the new resource.
     * @param resource      Resource instance for the new resource
     * @return the actual path that the server chose to use for our Resource
     * @throws org.wso2.registry.RegistryException
     *
     */
    public synchronized String put(String suggestedPath, Resource resource)
            throws RegistryException {

        suggestedPath = RegistryUtils.prepareGeneralPath(suggestedPath);

        RequestContext requestContext = new RequestContext(this);
        requestContext.setResourcePath(suggestedPath);
        requestContext.setResource(resource);

        if (!suggestedPath.equals(RegistryConstants.ROOT_PATH)) {
            handlerManager.putChild(requestContext);
        }

        handlerManager.put(requestContext);

        String actualPath = requestContext.getActualPath();

        if (!requestContext.isProcessingComplete()) {
            actualPath = suggestedPath;
            repository.put(suggestedPath, resource);

            // we are always creating versions for resources (files), if the resource has changed.
            if (!(resource instanceof Collection) &&
                    ((ResourceImpl) resource).isVersionableChange()) {
                createVersion(actualPath);
            }
        }

        logsDAO.addLog(suggestedPath, CurrentSession.getUser(), LogEntry.UPDATE, null);

        if (actualPath == null) {
            return suggestedPath;
        } else {
            return actualPath;
        }
    }

    public String importResource(String suggestedPath, String sourceURL, Resource metadata)
            throws RegistryException {

        suggestedPath = RegistryUtils.prepareGeneralPath(suggestedPath);

        RequestContext importChildContext = new RequestContext(this);
        importChildContext.setResourcePath(RegistryUtils.getParentPath(suggestedPath));
        handlerManager.importChild(importChildContext);

        RequestContext requestContext = new RequestContext(this);
        requestContext.setResourcePath(suggestedPath);
        requestContext.setSourceURL(sourceURL);
        requestContext.setResource(metadata);

        handlerManager.importResource(requestContext);
        String savedPath = requestContext.getActualPath();

        if (!requestContext.isProcessingComplete()) {
            savedPath = suggestedPath;
            repository.importResource(suggestedPath, sourceURL, metadata);
        }

        logsDAO.addLog(savedPath, CurrentSession.getUser(), LogEntry.UPDATE, null);

        return suggestedPath;
    }

    /**
     * Deletes the resource in the given path. If the path refers to a directory, all child
     * resources of the directory will also be deleted.
     *
     * @param path Path of the resource to be deleted.
     * @throws org.wso2.registry.RegistryException
     *          : if the user is not authorized
     */
    public synchronized void delete(String path) throws RegistryException {

        RequestContext requestContext = new RequestContext(this);
        requestContext.setResourcePath(path);

        handlerManager.delete(requestContext);

        if (!requestContext.isProcessingComplete()) {
            repository.delete(path);
        }

        logsDAO.addLog(path, CurrentSession.getUser(), LogEntry.DELETE_RESOURCE, null);
    }

    public String rename(String currentPath, String newName) throws RegistryException {

        String newPath  = repository.rename(currentPath, newName);
        logsDAO.addLog(currentPath, CurrentSession.getUser(), LogEntry.RENAME, newName);

        return newPath;
    }

    public String move(String currentPath, String newPath) throws RegistryException {

        String movedPath  = repository.move(currentPath, newPath);
        logsDAO.addLog(currentPath, CurrentSession.getUser(), LogEntry.MOVE, newPath);

        return movedPath;
    }

    public String copy(String sourcePath, String targetPath) throws RegistryException {

        String copiedPath  = repository.copy(sourcePath, targetPath);
        logsDAO.addLog(sourcePath, CurrentSession.getUser(), LogEntry.COPY, targetPath);

        return copiedPath;
    }

    public void createVersion(String path) throws RegistryException {

        versionRepository.createSnapshot(path);
    }

    public String[] getVersions(String path) throws RegistryException {

        return versionRepository.getVersions(path);

        //String[] versionPaths = null;
        //
        //Connection conn = getConnection();
        //
        //try {
        //    long artifactID = resourceDAO.getResourceID(path, conn);
        //    String[] versionIDs = resourceDAO.getVersionNumbers(artifactID, conn);
        //
        //    versionPaths = new String[versionIDs.length];
        //    for (int i = 0; i < versionIDs.length; i++) {
        //        versionPaths[i] = path + "?v=" + versionIDs[i];
        //    }
        //
        //} catch (SQLException e) {
        //
        //    try {
        //        conn.rollbackTransaction();
        //    } catch (SQLException e1) {
        //        e1.printStackTrace();
        //    }
        //    String msg = Messages.getMessage("could.not.found.versionpath", path, e.getMessage());
        //    log.error(msg, e);
        //    throw new RegistryException(msg, e);
        //
        //} finally {
        //    try {
        //        conn.close();
        //    } catch (SQLException ignore) {
        //        log.info("Nothing we can do here : Inside the finally block ");
        //    }
        //}
        //
        //return versionPaths;
    }

    public void restoreVersion(String versionPath) throws RegistryException {

        versionRepository.restoreVersion(versionPath);

        //String plainPath;
        //long versionNumber;
        //if (versionPath.indexOf("?v=") != -1) {
        //    String[] parts = versionPath.split("\\?v=");
        //    plainPath = parts[0];
        //    versionNumber = Long.parseLong(parts[1]);
        //} else {
        //    String msg = Messages.getMessage("not.a.valid.verion.path");
        //    log.error(msg);
        //    throw new RegistryException(msg);
        //}
        //
        //String userID = CurrentSession.getUser();
        //
        //Connection conn = getConnection();
        //
        //try {
        //    conn.setAutoCommit(false);
        //
        //    long artifactID = resourceDAO.getResourceID(plainPath, conn);
        //
        //    resourceDAO.restore(artifactID, versionNumber, conn);
        //
        //    logsDAO.addLog(
        //            artifactID, userID, LogEntry.RESTORE, Long.toString(versionNumber), conn);
        //
        //    conn.commit();
        //
        //} catch (Exception e) {
        //    String msg = Messages.getMessage("unable.to.restore", versionPath, e.getMessage());
        //    log.error(msg, e);
        //
        //    try {
        //        conn.rollbackTransaction();
        //    } catch (SQLException e1) {
        //        log.error(e1);
        //    }
        //
        //    throw new RegistryException(msg, e);
        //
        //} finally {
        //    try {
        //        conn.close();
        //    } catch (SQLException ignore) {
        //        log.info(Messages.getMessage("exception.closing.db"));
        //    }
        //}
    }

    ////////////////////////////////////////////////////////
    // Associations
    ////////////////////////////////////////////////////////
    public void addAssociation(String associationType, String sourcePath, String associationPath)
            throws RegistryException {

        String resourceID1 = resourceDAO.getResourceID(sourcePath);
        String resourceID2 = resourceDAO.getResourceID(associationPath);
        associationDAO.addAssociation(resourceID1, resourceID2, associationType);
    }

    public void removeAssociation(
            String associationType, String sourcePath, String associationPath)
            throws RegistryException {

        String resourceID1 = resourceDAO.getResourceID(sourcePath);
        String resourceID2 = resourceDAO.getResourceID(associationPath);
        associationDAO.removeAssociation(resourceID1, resourceID2, associationType);
    }

    public Association[] getAllAssociations(String resourcePath) throws RegistryException {

        return associationDAO.getAllAssociations(resourcePath);
    }

    public Association[] getAssociations(String resourcePath, String associationType)
            throws RegistryException {

        return associationDAO.getAllAssociationsForType(resourcePath, associationType);
    }

    public void addDependencies(String dependentPath, String[] dependencyPaths)
            throws RegistryException {

        throw new RegistryException("Associations are not yet implemented.");

        //VersionedPath versionedPath = RegistryUtils.getVersionedPath(dependentPath);
        //
        //Connection conn = null;
        //
        //try {
        //    conn = getConnection();
        //    long dependentID = resourceDAO.getResourceID(versionedPath.getPath(), conn);
        //
        //    if (versionedPath.getVersion() == -1) {
        //        versionedPath.setVersion(
        //                resourceDAO.getCurrentVersionNumber(versionedPath.getPath(), conn));
        //    }
        //
        //    resourceDAO.addDependencies(
        //            dependentID, versionedPath.getVersion(), dependencyPaths, conn);
        //
        //} catch (SQLException e) {
        //    String msg = "Failed to add dependencies for the resource: " + dependentPath;
        //    log.error(msg, e);
        //
        //    try {
        //        conn.rollbackTransaction();
        //    } catch (SQLException e1) {
        //        log.error(e1);
        //    }
        //
        //    throw new RegistryException(msg, e);
        //
        //} finally {
        //    if (conn != null) {
        //        try {
        //            conn.close();
        //        } catch (SQLException ignore) {
        //            log.info(Messages.getMessage("exception.closing.db"));
        //        }
        //    }
        //}
    }

    public DependencyChain[] getAllDependencies(String resourcePath) throws RegistryException {

        throw new RegistryException("Associations are not yet implemented.");

        //DependencyChain[] dependencies = null;
        //
        //Connection conn = null;
        //try {
        //    conn = getConnection();
        //
        //    dependencies = resourceDAO.getDependencies(resourcePath, conn);
        //
        //} catch (SQLException e) {
        //    String msg = "Failed to get dependencies of the resource: " + resourcePath;
        //    log.error(msg, e);
        //
        //    try {
        //        conn.rollbackTransaction();
        //    } catch (SQLException e1) {
        //        log.error(e1);
        //    }
        //
        //    throw new RegistryException(msg, e);
        //
        //} finally {
        //    if (conn != null) {
        //        try {
        //            conn.close();
        //        } catch (SQLException ignore) {
        //            log.info(Messages.getMessage("exception.closing.db"));
        //        }
        //    }
        //}
        //
        //return dependencies;
    }

    public DependentChain[] getAllDependents(String resourcePath) throws RegistryException {

        throw new RegistryException("Associations are not yet implemented.");

        //DependentChain[] dependents = null;
        //
        //Connection conn = null;
        //try {
        //    conn = getConnection();
        //
        //    dependents = resourceDAO.getDependents(resourcePath, conn);
        //
        //} catch (SQLException e) {
        //    String msg = "Failed to get dependents of the resource: " + resourcePath;
        //    log.error(msg, e);
        //
        //    try {
        //        conn.rollbackTransaction();
        //    } catch (SQLException e1) {
        //        log.error(e1);
        //    }
        //
        //    throw new RegistryException(msg, e);
        //
        //} finally {
        //    if (conn != null) {
        //        try {
        //            conn.close();
        //        } catch (SQLException ignore) {
        //            log.info(Messages.getMessage("exception.closing.db"));
        //        }
        //    }
        //}
        //
        //return dependents;
    }

    ////////////////////////////////////////////////////////
    // Tagging
    ////////////////////////////////////////////////////////

    /**
     * Applies the given tag to the resource in the given path. If the given tag is not defined in
     * the registry, it will be defined.
     *
     * @param resourcePath Path of the resource to be tagged.
     * @param tag          Tag. Any string can be used for the tag.
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public synchronized void applyTag(String resourcePath, String tag) throws RegistryException {

        // break the comma separated words into multiple tags
        String[] tags = tag.split(",");

        String resourceID;
        if ((resourceID = resourceDAO.getResourceID(resourcePath)) == null) {
            String msg = "Failed to apply tag " + tag + " on resource " + resourcePath +
                    ". Resource " + resourcePath + " does not exist.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        String userName = CurrentSession.getUser();

        for (int i = 0; i < tags.length; i++) {

            tags[i] = tags[i].trim();

            if (tags[i].length() == 0 || tags[i].equals(" ")) {
                continue;
            }

            if (tagsDAO.taggingExist(tags[i], resourceID, userName)) {
                // Already there, don't bother doing it again.
                return;
            }

            tagsDAO.addTagging(tags[i], resourceID, userName);
            logsDAO.addLog(resourcePath, userName, LogEntry.TAG, tags[i]);
        }
    }

    /**
     * Returns the paths of the resources, which are tagged using the given tag.
     *
     * @param tag Tag
     * @return Resource paths
     */
    public TaggedResourcePath[] getResourcePathsWithTag(String tag) throws RegistryException {

        // break the tags from spaces
        String[] tags = tag.trim().split(",");

        for (int i = 0; i < tags.length; i++) {
            tags[i] = tags[i].trim();
        }

        List taggedPaths = new ArrayList();

        List pathList = tagsDAO.getPathsWithAnyTag(tags);
        Iterator iPaths = pathList.iterator();
        while (iPaths.hasNext()) {
            String path = (String)iPaths.next();

            TaggedResourcePath taggedResourcePath = new TaggedResourcePath();
            taggedResourcePath.setResourcePath(path);

            for (String currentTag : tags) {
                long count = tagsDAO.getTagCount(path, currentTag);
                taggedResourcePath.addTagCount(currentTag, count);
            }

            taggedPaths.add(taggedResourcePath);
        }

        return (TaggedResourcePath[])taggedPaths
                .toArray(new TaggedResourcePath[taggedPaths.size()]);
    }

    /**
     * Returns all tags used for tagging the given resource.
     *
     * @param resourcePath Path of the resource
     * @return Tags
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public Tag[] getTags(String resourcePath) throws RegistryException {

        resourcePath = RegistryUtils.prepareGeneralPath(resourcePath);
        resourcePath = getCurrentPath(resourcePath);

        Tag[] tags = tagsDAO.getTagsWithCount(resourcePath);

        return tags;
    }

    public void removeTag(String path, String tag) throws RegistryException {

        path = RegistryUtils.prepareGeneralPath(path);

        String user = CurrentSession.getUser();

        boolean adminUser = false;

        // check is the user belongs to the admin role
        try {
            UserRealm userRealm = CurrentSession.getRealm();
            String[] roles = userRealm.getUserStoreReader().getUserRoles(user);
            if (RegistryUtils.containsString(RegistryConstants.ADMIN_ROLE, roles)) {
                adminUser = true;
            }

        } catch (UserStoreException e) {

            String msg = "Failed to get roles of the current user. " +
                    "User will be considered as non-admin user.\n" + e.getMessage();
            log.info(msg, e);
            adminUser = false;
        }

        // check if the user is the admin user
        if (RegistryConstants.ADMIN_USER.equals(user)) {
            adminUser = true;
        }

        String author = resourceDAO.getResourceAuthor(path);
        if (user.equals(author) || adminUser) {
            tagsDAO.removeTag(path, tag);
        } else {
            tagsDAO.removeTag(path, tag, user);
        }
    }

    ////////////////////////////////////////////////////////
    // Comments
    ////////////////////////////////////////////////////////

    /**
     * Adds a comment to a resource.
     *
     * @param resourcePath Path of the resource to add the comment.
     * @param comment      Comment instance for the new comment.
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public synchronized String addComment(String resourcePath, Comment comment)
            throws RegistryException {

        String userID = CurrentSession.getUser();

        String resourceID;
        if ((resourceID = resourceDAO.getResourceID(resourcePath)) == null) {
            String msg = "Failed to add comment " + comment.getText() + " to resource " +
                    resourcePath + ". Resource " + resourcePath + " does not exist.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        String commentPath = commentsDAO.addComment(resourceID, userID, comment);
        commentPath = resourcePath + ";comments:" + commentPath;
        logsDAO.addLog(
                resourcePath, userID, LogEntry.COMMENT, comment.getText());

        return commentPath;
    }


    public void editComment(String commentPath, String text) throws RegistryException {

        String[] parts = commentPath.split(";");
        String commentPart = parts[1];

        if (parts.length == 2 && commentPart.startsWith("comments:")) {
            String commentId = parts[1].substring(9);
            commentsDAO.updateComment(Long.parseLong(commentId), text);
        }
    }

    /**
     * Returns all comments for the given resource.
     *
     * @param resourcePath Path of the resource.
     * @return Array of Comment objects.
     */
    public Comment[] getComments(String resourcePath) throws RegistryException {

        resourcePath = getCurrentPath(resourcePath);
        Comment[] comments =  commentsDAO.getComments(resourcePath);

        return comments;
    }

    ////////////////////////////////////////////////////////
    // Ratings
    ////////////////////////////////////////////////////////

    /**
     * Rates the given resource.
     *
     * @param resourcePath Path of the resource.
     * @param rating       rating value.
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public synchronized void rateResource(String resourcePath, int rating)
            throws RegistryException {

        String userID = CurrentSession.getUser();

        String resourceID = resourceDAO.getResourceID(resourcePath);

        if (resourceID != null) {

            if (ratingsDAO.ratingExist(resourceID, userID)) {

                ratingsDAO.updateRating(resourceID, userID, rating);

                log.info("Updated the rating on the resource " +
                        resourcePath + " by user " + userID);
            } else {
                ratingsDAO.addRating(resourceID, userID, rating);
            }

            logsDAO.addLog(resourcePath,
                    userID, LogEntry.RATING, Integer.toString(rating));

        } else {
            String msg = Messages.getMessage("rate.on.null.artfact", resourcePath);
            log.error(msg);
            throw new RegistryException(msg);
        }
    }

    /**
     * Returns the average rating (from all users) for the given resource.
     *
     * @param resourcePath Path of the resource.
     * @return average rating.
     */
    public float getAverageRating(String resourcePath) throws RegistryException {

        float rating = ratingsDAO.getAverageRating(resourcePath);
        return rating;
    }

    public int getRating(String path, String userName) throws RegistryException {

        int rating = ratingsDAO.getRating(path, userName);
        return rating;
    }

    //////////////////////////////////////////////////////////////////////////////
    // Extensible searching API
    //////////////////////////////////////////////////////////////////////////////

    public Collection executeQuery(String path, Map parameters) throws RegistryException {

        Resource query = get(path);

        Collection resultCollection =  queryProcessorManager.executeQuery(query, parameters);

        return resultCollection;
    }


    private String getCurrentPath(String resourcePath) {
        String currentPath = resourcePath;
        if (resourcePath.indexOf("?") > 0) {
            currentPath = resourcePath.split("\\?")[0];
        } else if (resourcePath.indexOf(";") > 0) {
            currentPath = resourcePath.split("\\;")[0];
        }

        return currentPath;
    }

    public LogEntry[] getLogs(String resourcePath,
                              int action,
                              String userName,
                              Date from, Date to, boolean recentFirst) throws RegistryException {

        List logEntryList  =
                logsDAO.getLogs(resourcePath, action, userName, from, to, recentFirst);

        LogEntry[] logEntries = new LogEntry[logEntryList.size()];
        for (int i = 0; i < logEntryList.size(); i++) {
            logEntries[i] = (LogEntry)logEntryList.get(i);
        }

        return logEntries;
    }

    public void addAspect(String name, Aspect aspect) throws RegistryException {
        String msg = "Transactional registry does not support adding aspects.";
        log.error(msg);
        throw new RegistryException(msg);
    }

    public void associateAspect(String resourcePath, String name) throws RegistryException {
        //TODO need to do the security validation here
        Resource resource = get(resourcePath);
        Aspect aspect = getAspect(name);
        aspect.associate(resource, this);
        resource.addAspect(name);
        put(resource.getPath(), resource);
    }

    private Aspect getAspect(String name) throws RegistryException {
        Aspect aspect = aspects.get(name);
        if (aspect == null) {
            RegistryContext registryContext = RegistryContext.getRegistryContext();
            //RegistryContext registryContext = (RegistryContext)System.
            //        getProperties().get(RegistryConstants.REGISTRY_CONTEXT);
            aspect = registryContext.getAspect(name);
        }
        if (aspect == null) {
            throw new RegistryException(" No aspect object found for the name :"  + name);
        }
        return aspect;
    }

    //public void addAspect(String name, Aspect aspect) {
    //    aspects.put(name, aspect);
    //}

    public Aspect getResourceAspect(Resource resource, String aspectName)
            throws RegistryException {

        Aspect aspect = getAspect(aspectName);
        if (aspect == null) {
            throw new RegistryException("Aspect '" + aspectName + "' is not registered!");
        }

        // Confirm this aspect is associated with this Resource
        List<String> resourceAspects = resource.getAspects();
        if (resourceAspects == null || !resourceAspects.contains(aspectName)) {
            throw new RegistryException("Resource at '" + resource.getPath() +
                    "' not associated with aspect '" + aspectName + "'");
        }

        return aspect;
    }

    public void invokeAspect(String resourcePath, String aspectName, String action)
            throws RegistryException {

        Resource resource = get(resourcePath);

        Aspect aspect = getResourceAspect(resource, aspectName);
        RequestContext context = new RequestContext(this);
        context.setResource(resource);

//        List aspectNames = resource.getPropertyValues(Aspect.AVAILABLE_ASPECTS);
//        if (aspectNames == null) {
//            throw new RegistryException("No aspect are associated with the resource");
//        }

        aspect.transition(context, action);
        put(resourcePath, resource);
    }

    public String[] getAspectActions(String resourcePath, String aspectName)
            throws RegistryException {

        Resource resource = get(resourcePath);
        Aspect aspect = getResourceAspect(resource, aspectName);

        RequestContext context = new RequestContext(this);
        context.setResource(resource);
        return aspect.getAvailableActions(context);
    }
}
