/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc;

import org.wso2.registry.*;
import org.wso2.registry.jdbc.dao.ResourceDAO;
import org.wso2.registry.jdbc.dao.ResourceVersionDAO;
import org.wso2.registry.jdbc.dao.RestoreUtilDAO;
import org.wso2.registry.jdbc.dataobjects.ResourceDO;
import org.wso2.registry.utils.VersionedPath;
import org.wso2.registry.utils.RegistryUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.sql.DataSource;
import java.util.List;
import java.util.Iterator;
import java.util.ArrayList;
import java.io.InputStream;

public class VersionRepository {

    private static Log log = LogFactory.getLog(VersionRepository.class);

    private DataSource dataSource;

    private ResourceDAO resourceDAO = new ResourceDAO();
    private ResourceVersionDAO resourceVersionDAO = new ResourceVersionDAO();
    private RestoreUtilDAO restoreUtilDAO = new RestoreUtilDAO();

    public VersionRepository(DataSource dataSource) {
        this.dataSource = dataSource;
    }

    public void createSnapshot(String path) throws RegistryException {

        path = RegistryUtils.getPureResourcePath(path);

        String snapshotRootID = resourceDAO.getResourceID(path);

        long snapshotID = resourceVersionDAO.addSnapshot(snapshotRootID);
        versionSnapshotNetwork(snapshotID, snapshotRootID);
    }

    private void versionSnapshotNetwork(long snapshotID, String resourceID)
            throws RegistryException {

        int equivalentVersion = resourceDAO.getEquivalentVersion(resourceID);

        List <String> childIDs = resourceDAO.getChildIDs(resourceID);

        if (equivalentVersion == -1) {
            // resource has changed since the last version.

            ResourceImpl resourceImpl = resourceDAO.getResourceByID(resourceID);
            if (resourceImpl instanceof CollectionImpl) {
                resourceDAO.fillChildren((CollectionImpl) resourceImpl, -1, -1);
            }
            resourceDAO.fillResourceProperties(resourceImpl);

            // add the resource content to the content version table and get the ID
            // todo: if the resource content has not changed between two versions we can share the
            // todo: same content.
            resourceImpl.setDataSource(dataSource);
            InputStream contentStream = resourceImpl.getContentStream();
            String contentVersionID = resourceVersionDAO.addContentVersion(contentStream);

            int versionNumber = resourceVersionDAO.
                    addResourceVersion(resourceImpl, contentVersionID);
            resourceVersionDAO.addDependencies(resourceID, versionNumber, childIDs);
            resourceVersionDAO.addSnapshotResource(snapshotID, resourceID, versionNumber);
            resourceVersionDAO.addProperties(resourceImpl, versionNumber);
        }

        Iterator<String> childIterator = childIDs.iterator();
        while (childIterator.hasNext()) {

            String childID = childIterator.next();
            versionSnapshotNetwork(snapshotID, childID);
        }
    }

    public String[] getVersions(String resourcePath) throws RegistryException {

        resourcePath = RegistryUtils.getPureResourcePath(resourcePath);

        Long[] snapshotNumbers = resourceVersionDAO.getSnapshotIDs(resourcePath);

        List <String> versionPaths = new ArrayList();
        for (int i = 0; i < snapshotNumbers.length; i++) {
            String versionPath = resourcePath + ";version=" + snapshotNumbers[i];
            versionPaths.add(versionPath);
        }

        return versionPaths.toArray(new String[versionPaths.size()]);
    }

    public Resource get(VersionedPath versionedPath) throws RegistryException {

        String resourceID = resourceDAO.getResourceID(versionedPath.getPath());
        long snapshotID = versionedPath.getVersion();

        long version = resourceVersionDAO.getVersion(resourceID, snapshotID);

        if (version == -1) {
            String msg = "Resource " + versionedPath.getPath() +
                    " does not have a version " + versionedPath.getVersion();
            log.error(msg);
            throw new RegistryException(msg);
        }

        ResourceImpl resourceImpl = resourceVersionDAO.get(resourceID, snapshotID, version);
        resourceImpl.setDataSource(dataSource);

        return resourceImpl;
    }

    public void restoreVersion(String versionPath) throws RegistryException {

        VersionedPath versionedPath = RegistryUtils.getVersionedPath(versionPath);
        if (versionedPath.getVersion() == -1) {
            String msg = "Failed to restore resource. " +
                    versionedPath + " is not a valid version path.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        String versionedResourcePath = versionedPath.getPath();
        long snapshotID = versionedPath.getVersion();
        String resourceID = resourceDAO.getResourceID(versionedResourcePath);
        restoreSnapshotNetwork(snapshotID, resourceID);
    }

    private void restoreSnapshotNetwork(long snapshotID, String resourceID)
            throws RegistryException {

        // get the version number of resource for this snapshot
        // check if the resource exists in the resource table
        // if no,
        //      copy version to the resource table
        //      copy content to the content table and set the content ID of the resource
        //      copy all versioned properies to the properties table
        // if yes,
        //      check if that version number and equivalent version number are same
        //      if no,
        //          replace versionalble fields of the resource table
        //          delete current version content
        //          copy versioned content to content table, and set the content ID of the resource
        //          delete all properties of old resource and copy properties from the version
        //      end if
        // end if
        // get children of the version
        // get chledren of the current version
        // delete removed children
        // for each child
        //      call restoreSnapshotNetwork
        // end for

        ResourceDO resourceVersionDO =
                resourceVersionDAO.getResourceVersionDO(resourceID, snapshotID);
        ResourceDO resourceDO = resourceDAO.getResourceDO(resourceID);
        if (resourceDO != null) {

            if (resourceDO.getEquivalentVersion() != resourceVersionDO.getVersion()) {
                restoreUtilDAO.replaceContentVersion(resourceDO, resourceVersionDO);
                restoreUtilDAO.mergeResourceVersion(resourceDO, resourceVersionDO);
                restoreUtilDAO.replaceProperties(resourceID, resourceVersionDO.getVersion());
            }

        } else {

            String contentID = restoreUtilDAO.copyContentVersion(resourceDO, resourceVersionDO);
            restoreUtilDAO.copyResourceVersion(resourceVersionDO, contentID);
            restoreUtilDAO.copyProperties(resourceID, resourceVersionDO.getVersion());
        }

        List <String> versionedChildIDs =
                resourceVersionDAO.getChildIDs(resourceID, resourceVersionDO.getVersion());
        
        if (resourceDO != null) {
            List <String> childIDs = resourceDAO.getChildIDs(resourceID);
            Iterator <String> iChildIDs = childIDs.iterator();
            while (iChildIDs.hasNext()) {
                String childID = iChildIDs.next();
                if (!versionedChildIDs.contains(childID)) {
                    resourceDAO.delete(childID);
                }
            }
        }

        restoreUtilDAO.replaceDependencies(resourceID, versionedChildIDs);

        Iterator <String> iVersionedChildIDs = versionedChildIDs.iterator();
        while (iVersionedChildIDs.hasNext()) {
            restoreSnapshotNetwork(snapshotID, iVersionedChildIDs.next());
        }
    }
}
