/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */

package org.wso2.registry.jdbc;

import junit.framework.TestCase;
import org.wso2.registry.*;
import org.wso2.registry.jdbc.realm.RegistryRealm;
import org.wso2.registry.jdbc.utils.RegistryDataSource;

import javax.sql.DataSource;

public class VersionHandlingTest extends TestCase {

    protected static EmbeddedRegistry embeddedRegistry = null;
    protected static Registry registry = null;
    protected static RegistryRealm realm = null;

    public void setUp() throws RegistryException {
        if (registry == null) {
            DataSource dataSource = new RegistryDataSource(
                    "jdbc:derby://localhost:1527//home/chathura/test/version/db1;create=true",
                    "org.apache.derby.jdbc.ClientDriver", "admin", "admin");
            //embeddedRegistry = new EmbeddedRegistry(dataSource);
            embeddedRegistry = new InMemoryEmbeddedRegistry();
            registry = embeddedRegistry.getUserRegistry(
                    RegistryConstants.ADMIN_USER, RegistryConstants.ADMIN_PASSWORD);
        }
    }

    public void testCreateVersions() throws RegistryException {

        Resource r1 = registry.newResource();
        r1.setContent("some content");
        registry.put("/version/r1", r1);

        //registry.createVersion("/version/r1");

        String[] r1Versions = registry.getVersions("/version/r1");

        assertEquals("/version/r1 should have 1 version.", r1Versions.length, 1);

        Resource r1v2 = registry.get("/version/r1");
        r1v2.setContent("another content");
        registry.put("/version/r1", r1v2);

        //registry.createVersion("/version/r1");

        r1Versions = registry.getVersions("/version/r1");
        assertEquals("/version/r1 should have 2 version.", r1Versions.length, 2);
    }

    public void testResourceContentVersioning() throws RegistryException {

        Resource r1 = registry.newResource();
        r1.setContent("content 1");
        registry.put("/v2/r1", r1);

        Resource r12 = registry.get("/v2/r1");
        r12.setContent("content 2");
        registry.put("/v2/r1", r12);

        String[] r1Versions = registry.getVersions("/v2/r1");

        Resource r1vv1 = registry.get(r1Versions[0]);

        assertEquals("r1's first version's content should be 'content 1'",
                new String((byte[]) r1vv1.getContent()), "content 1");

        Resource r1vv2 = registry.get(r1Versions[1]);

        assertEquals("r1's second version's content should be 'content 2'",
                new String((byte[]) r1vv2.getContent()), "content 2");
    }

    public void testResourcePropertyVersioning() throws RegistryException {

        Resource r1 = registry.newResource();
        r1.setContent("content 1");
        r1.addProperty("p1", "v1");
        registry.put("/v2/r1", r1);

        Resource r1v2 = registry.get("/v2/r1");
        r1v2.addProperty("p2", "v2");
        registry.put("/v2/r1", r1v2);

        String[] r1Versions = registry.getVersions("/v2/r1");

        Resource r1vv1 = registry.get(r1Versions[0]);

        assertEquals("r1's first version should contain a property p1 with value v1",
                r1vv1.getProperty("p1"), "v1");

        Resource r1vv2 = registry.get(r1Versions[1]);

        assertEquals("r1's second version should contain a property p1 with value v1",
                r1vv2.getProperty("p1"), "v1");

        assertEquals("r1's second version should contain a property p2 with value v2",
                r1vv2.getProperty("p2"), "v2");
    }

    public void testSimpleCollectionVersioning() throws RegistryException {

        Collection c1 = registry.newCollection();
        registry.put("/v3/c1", c1);

        registry.createVersion("/v3/c1");

        Collection c2 = registry.newCollection();
        registry.put("/v3/c1/c2", c2);

        registry.createVersion("/v3/c1");

        Collection c3 = registry.newCollection();
        registry.put("/v3/c1/c3", c3);

        registry.createVersion("/v3/c1");

        Collection c4 = registry.newCollection();
        registry.put("/v3/c1/c2/c4", c4);

        registry.createVersion("/v3/c1");

        Collection c5 = registry.newCollection();
        registry.put("/v3/c1/c2/c5", c5);

        registry.createVersion("/v3/c1");

        String[] c1Versions = registry.getVersions("/v3/c1");

        Collection c1v1 = (Collection) registry.get(c1Versions[0]);
        Collection c1v2 = (Collection) registry.get(c1Versions[1]);
        Collection c1v3 = (Collection) registry.get(c1Versions[2]);
    }

    public void testResourceRestore() throws RegistryException {

        Resource r1 = registry.newResource();
        r1.setContent("content 1");
        registry.put("/test/v10/r1", r1);

        Resource r1e1 = registry.get("/test/v10/r1");
        r1e1.setContent("content 2");
        registry.put("/test/v10/r1", r1e1);

        String[] r1Versions = registry.getVersions("/test/v10/r1");
        registry.restoreVersion(r1Versions[0]);

        Resource r1r1 = registry.get("/test/v10/r1");

        assertEquals("Restored resource should have content 'content 1'",
                "content 1", new String((byte[]) r1r1.getContent()));
    }

    public void testSimpleCollectionRestore() throws RegistryException {

        //Collection c1 = registry.newCollection();
        //registry.put("/test/v11/c1", c1);
        //
        //registry.createVersion("/test/v11/c1");
        //
        //Resource r1 = registry.newResource();
        //r1.setContent("r1c1");
        //registry.put("/test/v11/c1/r1", r1);
        //
        //registry.createVersion("/test/v11/c1");
        //
        //Resource r2 = registry.newResource();
        //r2.setContent("r1c1");
        //registry.put("/test/v11/c1/r2", r2);
        //
        //registry.createVersion("/test/v11/c1");
        //
        //String[] c1Versions = registry.getVersions("/test/v11/c1");
        //
        //registry.restoreVersion(c1Versions[0]);
        //Collection c1r1 = (Collection) registry.get("/test/v11/c1");
        //assertEquals("version 1 of c1 should not have any children", 0, c1r1.getChildren().length);
    }

    //protected static Registry registry = null;
    //
    //public void setUp() {
    //    try {
    //        if (registry == null) {
    //            RegistryRealm registryRealm = new InMemoryRegistryRealm();
    //            registry = new InMemoryJDBCRegistry(registryRealm);
    //        }
    //    } catch (RegistryException e) {
    //        fail("Failed to initialize the registry.");
    //    }
    //}
    //
    //public void testResourceRestore() throws RegistryException {
    //    Resource r1 = registry.newResource();
    //    r1.setContent("Content1");
    //    registry.put("/wso2/wsas/v1/r1/resource1", r1);
    //    registry.put("/wso2/wsas/v1/resource1", r1);
    //    Resource esb = registry.newResource();
    //    esb.setContent("ESB content");
    //    registry.put("/wso2/esb/v2/resource1", esb);
    //    Resource resource1 = registry.get("/wso2/wsas/v1");
    //    String[] content1 = ((String[])resource1.getContent());
    //    assertEquals(2, content1.length);
    //    assertEquals("/wso2/wsas/v1/r1", content1[0]);
    //    assertEquals("/wso2/wsas/v1/resource1", content1[1]);
    //    String oldWSASVersion = registry.getVersions("/wso2/wsas")[0];
    //    registry.delete("/wso2/wsas/v1");
    //
    //    Resource r2 = registry.newResource();
    //    r2.setContent("Content2");
    //    registry.put("/wso2/wsas/v1/r2/resource2", r2);
    //    registry.put("/wso2/wsas/v1/resource2", r2);
    //    resource1 = registry.get("/wso2/wsas/v1");
    //    content1 = ((String[])resource1.getContent());
    //    assertEquals(2, content1.length);
    //
    //    assertEquals("/wso2/wsas/v1/r2", content1[0]);
    //    assertEquals("/wso2/wsas/v1/resource2", content1[1]);
    //
    //    registry.restoreVersion(oldWSASVersion);
    //
    //    resource1 = registry.get("/wso2/wsas/v1");
    //    content1 = ((String[])resource1.getContent());
    //    assertEquals(2, content1.length);
    //    assertEquals("/wso2/wsas/v1/r1", content1[0]);
    //    assertEquals("/wso2/wsas/v1/resource1", content1[1]);
    //
    //    Resource newresource = registry.newResource();
    //    newresource.setContent("newcontent");
    //    registry.put("/wso2/wsas/v1/r3", newresource);
    //    resource1 = registry.get("/wso2/wsas/v1");
    //    content1 = ((String[])resource1.getContent());
    //    assertEquals(3, content1.length);
    //    assertEquals("/wso2/wsas/v1/r1", content1[0]);
    //    assertEquals("/wso2/wsas/v1/resource1", content1[1]);
    //    assertEquals("/wso2/wsas/v1/r3", content1[2]);
    //
    //    registry.restoreVersion(oldWSASVersion);
    //    resource1 = registry.get("/wso2/wsas/v1");
    //    content1 = ((String[])resource1.getContent());
    //    assertEquals(2, content1.length);
    //    assertEquals("/wso2/wsas/v1/r1", content1[0]);
    //    assertEquals("/wso2/wsas/v1/resource1", content1[1]);
    //
    //}
    //
    //public void testResourceVersioning() throws Exception {
    //    String mashupString = "testMashup";
    //
    //    Resource mashup = registry.newResource();
    //    mashup.setContent(mashupString);
    //    registry.put("/wso2/mashup", mashup);
    //
    //    Resource wsas = registry.newResource();
    //    wsas.setContent("testWSAS".getBytes());
    //    registry.put("/wso2/wsa", wsas);
    //
    //    Resource myMashup = registry.get("/wso2/mashup");
    //    String content = new String((byte[])myMashup.getContent());
    //    assertNotNull(content);
    //    assertEquals(mashupString, content);
    //}
}
